#include "Plateau.hpp"

void Plateau::initRandom(void){ //initialise la fct random()
  srand((unsigned int)time(NULL));
}

int Plateau::random(int range){ //genere un nb aleatoire entre O et range-1
  return (rand() % range);
}

Plateau::Plateau(void){ //constructeur
  /*
    Avant de fabriquer le plateau proprement dit, il faut initialiser la fct random,
    preparer la liste de sauvegarde des coups de la partie,
    preciser qu'il n'y a pas encore de fichier de sauvegarde par defaut et
    fait par defaut commencer la partie par le joueur
  */
  initRandom();
  partie = new Liste_coups();
  fichier_courant = NULL;
  qui_commence = JOUEUR_COMMENCE;
  /*
    Pour fabriquer un plateau, on cree d'abord toutes ses cases en laissant leurs champs vides.
    Puis on cree (une et une seule fois) tous les quintuplets.
    On en profite pour remplir certains champs des cases.
    Enfin, on initialise la valeur des cases.
  */
  int i,j,k,l,m; //indices pour les boucles 'for'
  // --------------------------------
  // creation de toutes les cases du plateau
  for (i=0; i<N; i++){
    for (j=0; j<N; j++){
      jeu[i][j]=new Case(); //pointeur sur case
    }
  }
  // --------------------------------
  // On va creer tous les quintuplets en 4 etapes (horizontalement, verticalement
  // 1er diagonale, 2eme diagonale). En mm tps on remplie les champs 'cases_voisines'
  // et 'quintuplets' des cases qui les composent
  // rem: il se peut qu'on mette plusieurs fois la meme case ds la liste 'cases_voisines'
  // cependant cela ne pose pas de probleme et n'a aucune influence sur la vitesse d'execution
  Case* cases_quint[5];
  // 1er etape: balayage des quintuplets horizontaux (-)
  for (i=0; i<N; i++){
    for (j=0; j<N-4; j++){
      for (k=0; k<5; k++){
	cases_quint[k]=jeu[i][j+k];
      }
      // creation d'un quintuplet
      Quintuplet *quint_temp = new Quintuplet(cases_quint);
      // remplissage des champs des cases qui le composent
      for (l=0; l<5; l++){
	cases_quint[l]->ajoute_quintuplet(quint_temp); //Construit la liste de quintuplets
	for (m=0; m<5; m++){
	  if (m != l) cases_quint[l]->ajoute_case_voisine(cases_quint[m]); /*Construit la liste 
									     des cases voisines*/
	}
      }
    }
  }
  // 2eme etape: balayage des quintuplets verticaux (|)
  for (i=0; i<N; i++){
    for (j=0; j<N-4; j++){
      for (k=0; k<5; k++){
	cases_quint[k]=jeu[j+k][i];
      }
      // creation d'un quintuplet
      Quintuplet *quint_temp = new Quintuplet(cases_quint);
      // remplissage des champs des cases qui le composent
      for (l=0; l<5; l++){
	cases_quint[l]->ajoute_quintuplet(quint_temp); //Construit la liste de quintuplets
	for (m=0; m<5; m++){
	  if (m != l) cases_quint[l]->ajoute_case_voisine(cases_quint[m]); /*Construit la liste 
									     des cases voisines*/
	}
      }
    }
  }
  // 3eme etape: balayage des quintuplets sur la diagonale montante (\)
  for (i=0; i<N-4; i++){
    for (j=0; j<N-4; j++){
      for (k=0; k<5; k++){
	cases_quint[k]=jeu[i+k][j+k];
      }
      // creation d'un quintuplet
      Quintuplet *quint_temp = new Quintuplet(cases_quint);
      // remplissage des champs des cases qui le composent
      for (l=0; l<5; l++){
	cases_quint[l]->ajoute_quintuplet(quint_temp); //Construit la liste de quintuplets
	for (m=0; m<5; m++){
	  if (m != l) cases_quint[l]->ajoute_case_voisine(cases_quint[m]); /*Construit la liste 
									     des cases voisines*/
	}
      }
    }
  }
  // 4eme etape: balayage des quintuplets sur la diagonale montante (/)
  for (i=4; i<N; i++){
    for (j=0; j<N-4; j++){
      for (k=0; k<5; k++){
	cases_quint[k]=jeu[i-k][j+k];
      }
      // creation d'un quintuplet
      Quintuplet *quint_temp = new Quintuplet(cases_quint);
      // remplissage des champs des cases qui le composent
      for (l=0; l<5; l++){
	cases_quint[l]->ajoute_quintuplet(quint_temp); //Construit la liste de quintuplets
	for (m=0; m<5; m++){
	  if (m != l) cases_quint[l]->ajoute_case_voisine(cases_quint[m]); /*Construit la liste 
									     des cases voisines*/
	}
      }
    }
  }
  // --------------------------------
  // Remplissage de la valeur des cases
  for (i=0; i<N; i++){
    for (j=0; j<N; j++){
      jeu[i][j]->calcule_valeur();
    }
  }
  affiche();
}

int Plateau::annuler_coup(Coup* coup){ /* annule un coup, le stocke ds 'coup' et retourne
					  0 en cas de reussite
					  -1 en cas d'echec (aucun coup a annuler) */
  if ( partie->lire_dernier_coup(coup) == -1){ //aucun coup a annuler
    return -1; 
  }
  else{ //si il reste au moins un coup a annuler
    //on met la case de coord (_x,_y) a VIDE et on recalcule la valeur des cases voisines
    jeu[coup->_y][coup->_x]->mise_a_jour(VIDE); //on recalcule la valeur des cases voisines
    partie->effacer_dernier_coup(); //on retire ce coup de la liste des coups
    coup->etat = VIDE; //on retourne les coord de ce coup en precisant que la cases est maintenant VIDE
    return 0;
  }
}

int Plateau::jouer_manuel(int j, int i){ /*Le joueur met un pion sur la case de la colonne j
					   et de la ligne i. La fct retourne:
					   0 si tout s'est bien passe
					   1 si le joueur a gagne
					   -1 en cas d'erreur
					   -2 si la case est deja prise */
  //rem: pas besoin de verifier si les coord sont bien ds la plateau grace a l'interface graphique
  int code = jeu[i][j]->mise_a_jour(CROIX); //le joueur a les croix et l'ordi les ronds
  if (code==0){ //tt s'est bien passe -> on sauvegarde le coup pour pouvoir l'annuler ulterieurement
    Coup _coup;
    _coup._x = j;
    _coup._y = i;
    _coup.etat = CROIX;
    partie->ajouter_coup(_coup);
  }
  return code;
}


int Plateau::jouer_automatique(int *x, int *y){ /* L'ordi selectionne une case de plus grande valeur, joue dedans
						   passe ses coord ds les var passes en param et retourne:
						   0 si tout s'est passe normalement
						   2 si l'ordi a gagne
						   -1 en cas d'erreur */
  int i, j, valeur_cur;
  int valeur_max = 0, occurances = 0;
  Case* tab_cases[N*N];
  int tab_i[N*N];
  int tab_j[N*N];
  //On liste ds tab_cases les 'occurances' cases de plus gde valeur
  for (i=0; i<N; i++){
    for (j=0; j<N; j++){
      if (jeu[i][j]->retourne_etat()==VIDE){ //l'ordi ne joue que sur les cases vides
	valeur_cur = jeu[i][j]->retourne_valeur();
	if (valeur_cur > valeur_max){
	  valeur_max = valeur_cur;
	  occurances = 1;
	  tab_cases[0] = jeu[i][j];
	  tab_i[0] = i;
	  tab_j[0] = j;
	}
	if (valeur_cur == valeur_max){
	  tab_cases[occurances] = jeu[i][j];
	  tab_i[occurances] = i;
	  tab_j[occurances] = j;
	  occurances ++;
	}
      }
    }
  }
  //On choisit une de ces cases aleatoirement
  int aleat = random(occurances);
  //L'ordi joue dessus
  switch (tab_cases[aleat]->mise_a_jour(ROND)){
  case 0: //tt s'est passe normalement -> on ajoute le coup a la liste et on retourne les coord
    Coup _coup;
    _coup._x = tab_j[aleat]; //correspond au j (axe des abscysses) joue par l'ordi
    _coup._y = tab_i[aleat]; //correspond au i joue par l'ordi
    _coup.etat = ROND;
    partie->ajouter_coup(_coup); //ajoute le coup a la liste de coups
    *x = tab_j[aleat]; 
    *y = tab_i[aleat]; //retourne les coord ds les var passes en param
    return 0;
    break;
  case 2: //l'ordi a gagne
    _coup._x = tab_j[aleat]; //correspond au j (axe des abscysses) joue par l'ordi
    _coup._y = tab_i[aleat]; //correspond au i joue par l'ordi
    _coup.etat = ROND;
    partie->ajouter_coup(_coup); //ajoute le coup a la liste de coups
    *x = tab_j[aleat]; 
    *y = tab_i[aleat]; //retourne les coord ds les var passes en param
    return 2;
    break;
  default: //erreur
    return -1;
  }
}

void Plateau::restart(void){
  //on efface les coups de la partie precedante
  while (partie->non_vide()){
    partie->effacer_dernier_coup();
  }
  //on vide toutes les cases
  int i,j;
  for (i=0; i<N; i++){
    for (j=0; j<N; j++){
      jeu[i][j]->reset();
    }
  }
  for (i=0; i<N; i++){
    for (j=0; j<N; j++){
      jeu[i][j]->reinitialise();
    }
  }
  //pour eviter de sauver cette nouvelle partie sur une ancienne sans le vouloir:
  if (fichier_courant != NULL) delete fichier_courant;
  fichier_courant = NULL;
}

char* Plateau::sauvegarder(char* fichier){ /* sauvegarde la partie ds fichier si celui-ci est precise
					    ou ds fichier_courant si fichier=NULL et retourne
					    'fichier_courant' en cas de succes
					    NULL en cas d'echec*/
  if ( (fichier==NULL) && (fichier_courant==NULL) ) //aucun fichier de sauvegarde precise
    return NULL;
  else{
    if (fichier == NULL) fichier=fichier_courant; //on sauvegarde ds le fichier courant
    else{ //on sauvegarde ds le fichier passe en param
      if (fichier_courant != NULL) delete fichier_courant;
      fichier_courant = fichier; //le fichier courant devient le fichier passe en param
    }
    FILE* fic;
    if ( (fic=fopen(fichier,"w")) == NULL){
      perror("fopen");
      exit(EXIT_FAILURE);
    }
    partie->sauvegarder(fic);
    fclose(fic);
    return fichier_courant;
  }
}  

int Plateau::charger(char* fichier, int tab_coups[N][N]){ /* charge une partie sauvegardee et retourne
				       1 en cas de succes -> c'est au joueur de jouer
				       2 en cas de succes -> c'est a l'ordi de jouer
				       -1 en cas d'echec*/
  if ( fichier == NULL ) //aucun fichier a charger precise
    return -1;
  else{ //on charge la partie selectionnee
    /* On ouvre et teste le fichier de sauvegarde */
    FILE* fic;
    //on teste si le fichier selectionne existe bien
    if ( (fic=fopen(fichier,"r")) == NULL){
      fclose(fic);
      return -1;
    }
    //on teste si le fichier est bien un fichier et s'il contient une sauvegarde de Gomoku
    char* test = new char[7];;
    fread(test,sizeof(char),(size_t)6,fic);
    if (strncmp(test,"GOMOKU",6) != 0){
      fclose(fic);
      return -1;
    }
    /* On ecrase la partie courante */
    //on vide toutes les cases du plateau
    int i,j;
    for (i=0; i<N; i++){
      for (j=0; j<N; j++){
	jeu[i][j]->reset();
      }
    }
    //on vide la liste des coups
    partie->vider();
    /* Chargement de la partie */
    //le fichier_courant devient le fichier passe en parametre
    if ( fichier_courant != NULL) delete fichier_courant;
    fichier_courant=fichier; 
    //la partie est chargee
    Coup _coup;
    while (fread(&_coup,sizeof(Coup),(size_t)1,fic) == 1){
      partie->ajouter_coup(_coup); //on remplit la liste des coups 1 par 1
      jeu[_coup._y][_coup._x]->mise_a_jour(_coup.etat); //on met a jour la case
      tab_coups[_coup._y][_coup._x] = _coup.etat; //servira a mettre a jour l'interface graphique
    }
    fclose(fic);
    //determine  a qui c'est le tour de jouer
    /*rem: en pratique c'est toujours au joueur de jouer car l'ordi joue toujours juste apres lui
      mais cela permet un certain controle et cette fct serait utile si on voulait ameliorer
      le pgm et faire jouer 2 joueurs ensemble */
    if (_coup.etat == CROIX) return 2;
    else return 1;
  }
}

void Plateau::change_qui_commence(void){ /* change celui qui doit commencer la partie
					    JOUEUR_COMMENCE(=0) -> ORDI_COMMENCE(=1)
					    ORDI_COMMENCE(=1) -> JOUEUR_COMMENCE(=0) */
  qui_commence = (qui_commence+1)%2;
}

int Plateau::dit_qui_commence(void){ //retourne qui doit commencer la partie
  return qui_commence;
}

// Pour le debuggage:
void Plateau::affiche(void){ //affiche la valeur des cases du plateau sur la ligne de commande
  int i,j;
  for (i=-1; i<N; i++){
    for (j=-1; j<N; j++){
      if ((i == -1)){
	cout.width(7);
	if ((j == -1)) cout << " ";
	else{
	  cout.width(7);
	  cout << j;
	  cout << "  ";
	}
      }
      else{
	if ((j == -1)){
	  cout.width(7);
	  cout << i << "|";
	}
	else jeu[i][j]->affiche();
      }
    }
    cout << endl;
  }
}
